;;; The contents of this directory structure is subject to the following
;;; copyright notice:
;;;
;;;	Copyright (c) 2006, 2007 Andrew A. Berlin
;;;
;;;	This program is free software; you can redistribute it and/or
;;;	modify it under the terms of the GNU General Public License as
;;;	published by the Free Software Foundation; either version 2 of
;;;	the License, or (at your option) any later version.
;;;
;;;	This program is distributed in the hope that it will be
;;;	useful, but WITHOUT ANY WARRANTY; without even the implied
;;;	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;;;	PURPOSE.  See the GNU General Public License for more details.
;;;
;;;	You should have received a copy of the GNU General Public
;;;	License along with this program; if not, write to the Free
;;;	Software Foundation, Inc., 59 Temple Place - Suite 330,
;;;	Boston, MA 02111-1307, USA.

#|
;;usage:
(define-world 'myworld
  (topology simple-linear-fallopian-tube)
  (probe 'main-probe
	 (initial-location (make-3point 3 4 5))
	 (initial-velocity (make-3point 10 21 0))
	 (magnet magnet-Gaussboys-1mmx4mm))
  (actuator 'actuator1
	    (location (make-3point 1 2 3))
	    (orientation-target (make-3point  9 10 11))
	    (polarity 'positive)
	    (magnet magnet-Gaussboys-large-37mmx37mm))
)
|#	    

(define (process-define-world-declaration decl)
  (cond ((pair? decl)
	 (let ((decl-type (car decl)))
	   (cond ((or (eq? decl-type 'probe) ;;check for named forms
		      (eq? decl-type 'actuator))
		  `(list ',(car decl) ,(cadr decl) ;;name
			 ,@(map process-define-world-declaration (cddr decl))))
		 ((member decl-type
			  '(topology
			    initial-location
			    initial-velocity
			    location
			    orientation-target
			    polarity
			    magnet))
		  `(list ',(car decl) ,@(cdr decl)))
		 (else
		  decl))))
	(else
	 decl)))

(define-syntax define-world
  (sc-macro-transformer (lambda (exp env) 
			  `(define-world-parser ,(cadr exp)
			     ,@(let loop ((declarations (cddr exp)))
				 (if (null? declarations)
				     '()
				     (cons (process-define-world-declaration (car declarations))
					   (loop (cdr declarations)))))))))


(define (define-world-parser name . args)
  (define (define-world-get-topology args)
    (let ((entry (assoc 'topology args)))
      (if entry
	  (second entry)
	  (error "Define-world: world must include a topology" name))))
  (let* ((topology (define-world-get-topology args))
	 (world (make-new-world topology name)))
    ;;Probe
    (define (define-world-process-probe args)
      (if (not (= (length args) 4))
	  (error "define-world-process-probe requires 4 arguments: name, initial-location, initial-velocity, and magnet" (length args)))
      (let* ((name (first args))
	     (location-entry (assoc 'initial-location (cdr args)))
	     (location (second location-entry))
	     (velocity-entry (assoc 'initial-velocity (cdr args)))
	     (velocity (second velocity-entry))
	     (magnet-entry (assoc 'magnet (cdr args)))
	     (magnet (second magnet-entry)))
	(make-probe-and-add-to-world! world
				      name
				      location
				      velocity
				      magnet)))
    ;;Actuators
    (define (define-world-process-actuator args)
      (if (not (= (length args) 5))
	  (error "define-world-process-actuator requires 5 arguments: name, location, orientation-target, polarity, and magnet" (length args)))
      (let* ((name (first args))
	     (location-entry (assoc 'location (cdr args)))
	     (location (second location-entry))
	     (orientation-target-entry (assoc 'orientation-target (cdr args)))
	     (orientation-target (second orientation-target-entry))
	     (polarity-entry (assoc 'polarity (cdr args)))
	     (polarity (second polarity-entry))
	     (magnet-entry (assoc 'magnet (cdr args)))
	     (magnet (second magnet-entry)))
	(if (not (member polarity '(positive negative)))
	    (error "Invalid polarity specified" polarity))
	(add-actuator-to-world!
	 world
	 (make-custom-actuator name
			       location
			       orientation-target
			       (if (eq? polarity 'negative)
				   #t ;;invert polarity
				   #f) ;;do not invert polarity
			       magnet))))

    ;;Dispatch loop
    (let loop ((commands args))
      (if (null? commands)
	  (begin (initialize-world! world)
		 world)
	  (let* ((command (car commands))
		 (keyword (car command)))
	    (cond ((eq? keyword 'topology)
		   (if (not (eq? (second command) (world-topology world)))
		       (error "World can contain only one fallopian tube topology" topology (world-topology world))))
		  ((eq? keyword 'probe)
		   (define-world-process-probe (cdr command)))
		  ((eq? keyword 'actuator)
		   (define-world-process-actuator (cdr command)))
		  (else
		   (error "Define-world: Unknown keyword" keyword)))
	    (loop (cdr commands)))))))

(define (initialize-world! world #!optional create-probes! create-actuators!)
  (if (not (default-object? create-probes!))
      (create-probes! world))
  (if (not (default-object? create-actuators!))
      (create-actuators! world))
  ;;add more here if needed, for instance this is where electromagnets would go
  (set-world-latest-state! world (world-initial-state world))
  (set-world-state-history! world (list (world-initial-state world)))
  )