;;; The contents of this directory structure is subject to the following
;;; copyright notice:
;;;
;;;	Copyright (c) 2006, 2007 Andrew A. Berlin
;;;
;;;	This program is free software; you can redistribute it and/or
;;;	modify it under the terms of the GNU General Public License as
;;;	published by the Free Software Foundation; either version 2 of
;;;	the License, or (at your option) any later version.
;;;
;;;	This program is distributed in the hope that it will be
;;;	useful, but WITHOUT ANY WARRANTY; without even the implied
;;;	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;;;	PURPOSE.  See the GNU General Public License for more details.
;;;
;;;	You should have received a copy of the GNU General Public
;;;	License along with this program; if not, write to the Free
;;;	Software Foundation, Inc., 59 Temple Place - Suite 330,
;;;	Boston, MA 02111-1307, USA.

(declare (usual-integrations))


;;A torque-and-forces-collection is a set of quantities that can be added together
;;i.e. they apply to the center of mass (or around the center of mass) of the probe
;;The raw forces of attraction acting on each monopole are NOT suitable for inclusion here,
;;as they require more complex geometric manipulation to shift into the coordinate
;;frame of the center of mass.
(define-structure torque-and-forces-collection
  Fpropulsive-mag
  Ftranslational
  Torque)

(define (compute-tf-collection world probe state pos-offset)
  (let ((tfs 
	 (map (lambda (actuator)
		(compute-torque-and-forces-due-to-actuator world actuator probe state pos-offset))
	      (world-actuators world))))
    (apply torque-and-forces-collection-+ tfs)))
	
(define (torque-and-forces-collection-+ tfc1 . rest)
  (if (null? rest)
      tfc1
      (let ((tfc2 (car rest))
	    (remaining (cdr rest)))
	(cond ((null? remaining)
	       (make-torque-and-forces-collection
		(+ (torque-and-forces-collection-Fpropulsive-mag tfc1)
		   (torque-and-forces-collection-Fpropulsive-mag tfc2))
		(3point-+ (torque-and-forces-collection-Ftranslational tfc1)
			  (torque-and-forces-collection-Ftranslational tfc2))
		(3point-+ (torque-and-forces-collection-Torque tfc1)
			  (torque-and-forces-collection-Torque tfc2))))
	      (else
	       (apply torque-and-forces-collection-+
		      (cons (torque-and-forces-collection-+ tfc1 tfc2)
			    remaining)))
	      ))))
  
(define (compute-torque-and-forces-due-to-actuator world actuator probe state pos-offset)
  (let* ((tfc-m1 (compute-torque-and-forces-on-probe-due-to-monopole world (actuator-monopole1 actuator) probe state pos-offset))
	 (tfc-m2 (compute-torque-and-forces-on-probe-due-to-monopole world (actuator-monopole2 actuator) probe state pos-offset)))
    (torque-and-forces-collection-+ tfc-m1 tfc-m2)))

(define (compute-torque-and-forces-due-to-actuating-probe world actuating-probe probe state pos-offset)
  (error "Unused")
  (if (eq? actuating-probe probe)
      (error "Actuating probe and motion probe must be different" (list actuating-probe probe state)))
  (let* ((probe-monopoles (make-probe-monopoles world probe state pos-offset))

	 (actuating-probe-monopoles (make-probe-monopoles world actuating-probe state (make-3point 0 0 0)))
	 (a-m1 (first actuating-probe-monopoles))
	 (a-m2 (second actuating-probe-monopoles))

	 (tfc-m1 (compute-torque-and-forces-on-probe-due-to-monopole world a-m1 probe state pos-offset))
	 (tfc-m2 (compute-torque-and-forces-on-probe-due-to-monopole world a-m2 probe state pos-offset)))
    
    (torque-and-forces-collection-+ tfc-m1 tfc-m2)))

(define *about-=-threshold* 1e-12)
(define (about-= a b)
  (if (3point? a)
      (and (about-= (3point-x a) (3point-x b))
	   (about-= (3point-y a) (3point-y b))
	   (about-= (3point-z a) (3point-z b)))
      (< (abs (- a b)) *about-=-threshold*)))

(define (compute-torque-and-forces-on-probe-due-to-monopole world monopole probe state pos-offset)
  (let* ((probe-position-alpha (vector-ref state 
					   (probe-position-reference probe)))
	 (probe-monopoles (make-probe-monopoles world probe state pos-offset))
	 (p-m1 (first probe-monopoles))
	 (p-m2 (second probe-monopoles))

	 ;;compute force acting on each probe monopole due to actuator monopole
	 (force-m1 (total-monopole-monopole-force p-m1 monopole))
	 (force-m2 (total-monopole-monopole-force p-m2 monopole))

	 (orientation (compute-probe-orientation world probe-position-alpha))

	 ;;compute component of force that is propulsive
	 (propulsive-force-mag-m1 (3point-dot-product force-m1 orientation))
	 (propulsive-force-mag-m2 (3point-dot-product force-m2 orientation))
	 (propulsive-force-m1 (scale-3point propulsive-force-mag-m1 orientation))
	 (propulsive-force-m2 (scale-3point propulsive-force-mag-m2 orientation))
	 (propulsive-force-total-mag (+ propulsive-force-mag-m1
					propulsive-force-mag-m2))

	 ;;compute component of force that is non-propulsive
	 (nonpropulsive-force-m1 (3point-- force-m1
					   propulsive-force-m1))
	 (nonpropulsive-force-m2 (3point-- force-m2
					   propulsive-force-m2))

	 (halflen (dipole-halflen (magnet-dipole (probe-magnet probe))))

	 ;;model the system in which F1 is non-propulsive force applied at m1 and F2 is non-propulsive force applied at m2
	 ;;as  F1 = Ftranslational/2 + Frotational/2
	 ;;and F2 = Ftranslational/2 - Frotational/2

	 (Ftranslational (3point-+ nonpropulsive-force-m1
				   nonpropulsive-force-m2))

	 (Frotational (3point-- nonpropulsive-force-m1
				nonpropulsive-force-m2))

	 ;;vector from center of probe to probe monopole1
	 (m1-r (scale-3point halflen orientation)) 
	 ;;vector from center of probe to probe monopole2
	 (m2-r (scale-3point (- halflen) orientation)) 

	 (Torque (3point-cross-product m1-r Frotational))
	 ;;torque should yield same result as:
	 (torque-check (3point-+ (3point-cross-product m1-r force-m1)
				 (3point-cross-product m2-r force-m2)))
	 )

    ;;double check the calculations - confidence builder
    (if (not (about-= Torque torque-check))
	(error "Torque check had a problem" (list torque torque-check)))

    ;;return result
    (make-torque-and-forces-collection propulsive-force-total-mag Ftranslational Torque)))

